import 'dart:convert';
import 'dart:typed_data';

import 'package:asn1lib/asn1lib.dart';
import 'package:pointycastle/api.dart';
import 'package:pointycastle/asymmetric/api.dart';


decryptByPublicKey(String data, PublicKey publicKey) {
  try {
//      var keyParameter = () => PrivateKeyParameter<RSAPrivateKey>(privateKey);
    var keyParameter = () => PublicKeyParameter<RSAPublicKey>(publicKey);
    AsymmetricBlockCipher cipher = AsymmetricBlockCipher("RSA/PKCS1");
    cipher.reset();
    cipher.init(false, keyParameter());
    Uint8List bconv = Base64Decoder().convert(data);
    final decrypted = cipher.process(bconv);
    String res = String.fromCharCodes(decrypted);
    return res;
  } catch (e) {
    print(e.toString());
  }
}

encryptByPrivateKey(String data, PrivateKey privateKey) {
  try {
//      var keyParameter = () => PublicKeyParameter<RSAPublicKey>(publicKey);
    var keyParameter = () => PrivateKeyParameter<RSAPrivateKey>(privateKey);
    AsymmetricBlockCipher cipher = AsymmetricBlockCipher("RSA/PKCS1");
    cipher.reset();
    cipher.init(true, keyParameter());
    Uint8List encryptResult = cipher.process(utf8.encode(data) as Uint8List);
    String encrypted = Base64Encoder().convert(encryptResult);
    return encrypted;
  } catch (e) {
    print(e.toString());
  }
}
decryptByPrivateKey(String data,RSAPrivateKey privateKey) {
  try {
    var keyParameter = () => PrivateKeyParameter<RSAPrivateKey>(privateKey);
//      var keyParameter = () => PublicKeyParameter<RSAPublicKey>(publicKey);
    AsymmetricBlockCipher cipher = AsymmetricBlockCipher("RSA/PKCS1");
    cipher.reset();
    cipher.init(false, keyParameter());
    Uint8List bconv = Base64Decoder().convert(data);
    final decrypted = cipher.process(bconv);
    String res = String.fromCharCodes(decrypted);
    return res;
  } catch (e) {
    print(e.toString());
  }
}
RSAAsymmetricKey parse(String key) {
  final rows = key.split('\n'); // LF-only, this could be a problem
  final header = rows.first;
  if (header == '-----BEGIN RSA PUBLIC KEY-----') {
    return _parsePublic(_parseSequence(rows));
  }

  if (header == '-----BEGIN PUBLIC KEY-----') {
    return _parsePublic(_pkcs8PublicSequence(_parseSequence(rows)));
  }

  if (header == '-----BEGIN RSA PRIVATE KEY-----') {
    return _parsePrivate(_parseSequence(rows));
  }

  if (header == '-----BEGIN PRIVATE KEY-----') {
    return _parsePrivate(_pkcs8PrivateSequence(_parseSequence(rows)));
  }
  // NOTE: Should we throw an exception?
  return null!;
}
RSAAsymmetricKey _parsePublic(ASN1Sequence sequence) {
  final modulus = (sequence.elements[0] as ASN1Integer).valueAsBigInteger;
  final exponent = (sequence.elements[1] as ASN1Integer).valueAsBigInteger;

  return RSAPublicKey(modulus!, exponent!);
}

RSAAsymmetricKey _parsePrivate(ASN1Sequence sequence) {
  final modulus = (sequence.elements[1] as ASN1Integer).valueAsBigInteger;
  final exponent = (sequence.elements[3] as ASN1Integer).valueAsBigInteger;
  final p = (sequence.elements[4] as ASN1Integer).valueAsBigInteger;
  final q = (sequence.elements[5] as ASN1Integer).valueAsBigInteger;

  return RSAPrivateKey(modulus!, exponent!, p, q);
}

ASN1Sequence _parseSequence(List<String> rows) {
  final keyText = rows
      .skipWhile((row) => row.startsWith('-----BEGIN'))
      .takeWhile((row) => !row.startsWith('-----END'))
      .map((row) => row.trim())
      .join('');

  final keyBytes = Uint8List.fromList(base64.decode(keyText));
  final asn1Parser = ASN1Parser(keyBytes);

  return asn1Parser.nextObject() as ASN1Sequence;
}

ASN1Sequence _pkcs8PublicSequence(ASN1Sequence sequence) {
  final ASN1BitString bitString = sequence.elements[1] as ASN1BitString;
  final bytes = bitString.valueBytes().sublist(1);
  final parser = ASN1Parser(Uint8List.fromList(bytes));

  return parser.nextObject() as ASN1Sequence;
}

ASN1Sequence _pkcs8PrivateSequence(ASN1Sequence sequence) {
  final ASN1BitString bitString = sequence.elements[2] as ASN1BitString;
  final bytes = bitString.valueBytes();
  final parser = ASN1Parser(bytes);

  return parser.nextObject() as ASN1Sequence;
}


